import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.DialogBoxClass;
import win32frame.PushButton;
import win32frame.TextBox;
import std.stdio;
import std.string;
import std.utf;

import HSVwheel;
import RGBslider;

COLORREF RGBA(uint r, uint g, uint b, uint a) {
  return cast(COLORREF)(r | (g << 8) | (b << 16) | (a<<24));
}

enum CMD{
  HSV = 0x100,
  RGB    ,
  OK     ,
  CANCEL
} ;

class ColorSelectDialog : DialogBoxClass {
  HSVwheel    hsvwheel  ;
  RGBslider   rgbslider ;

  PushButton  buttonOK  ;
  PushButton  buttonCAN ;
  HFONT       hfont  ;

  HDC         WinDC;
  HDC         WinDC_back;
  HBITMAP     bmp_back  ;
  BITMAPINFO  bi ;
  uint*       bgmap ;
  
  bool        enbAlpha=false ;
  uint        cur_color ;
  uint        def_color ;
  int         dlgstate  ;

  const int   defaultWidth =200 ;
  const int   defaultHeight=300 ;
  const DWORD defaultStyle = WS_VISIBLE | WS_SYSMENU ;
  const int   dispcolWidth = 32 ;
  const int   dispcolHeight=150 ;

  this(string name, string ClassName) {
    UnregisterClass( toUTFz!(wchar*)(ClassName), GetModuleHandle(null) ) ;
    winc.cbSize         = WNDCLASSEX.sizeof;
    winc.style          = 0;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.hIcon          = null ;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.hbrBackground  = CreateSolidBrush(GetSysColor(COLOR_BTNFACE)) ;
    winc.lpszMenuName   = null;
    winc.lpszClassName  = toUTFz!(wchar*)(ClassName) ;
    super(name) ;
  }

  int Create( int x, int y, HWND phwnd, HMENU hmenu){
    return( create(defaultStyle,x,y,defaultWidth,defaultHeight,phwnd,hmenu) ) ;
  }
  
  override LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      DeleteObject(hfont);
      DeleteDC(WinDC_back);
      DeleteObject(bmp_back);
      ReleaseDC(hwnd, WinDC);
      break ;
    case WM_CREATE:
      hfont = CreateFont(
			 12 , 0 , 0 , 0 , FW_REGULAR , FALSE , FALSE , FALSE ,
			 DEFAULT_CHARSET, OUT_DEFAULT_PRECIS ,
			 CLIP_DEFAULT_PRECIS , DEFAULT_QUALITY , 
			 FIXED_PITCH, null
			 );
      
      bi.bmiHeader.biSize             = BITMAPINFOHEADER.sizeof;
      bi.bmiHeader.biWidth            =  dispcolWidth  ;
      bi.bmiHeader.biHeight           = -dispcolHeight ;
      bi.bmiHeader.biPlanes           = 1;
      bi.bmiHeader.biBitCount         = 32;
      bi.bmiHeader.biCompression      = BI_RGB;
      bi.bmiHeader.biSizeImage        = 0;
      bi.bmiHeader.biXPelsPerMeter    = 0;
      bi.bmiHeader.biYPelsPerMeter    = 0;
      bi.bmiHeader.biClrUsed          = 1;
      bi.bmiHeader.biClrImportant     = 0;
      
      WinDC=GetDC(hwnd);
      WinDC_back=CreateCompatibleDC(WinDC);
      bmp_back  =CreateDIBSection( WinDC, &bi, DIB_RGB_COLORS, cast(void**)(&bgmap), null, 0 );
      SelectObject(WinDC_back, bmp_back);
      fillColor(true) ;

      hsvwheel = new HSVwheel("HSVwheel","HSVwheel") ;
      hsvwheel.setStyle(WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE) ;
      hsvwheel.setPosition(0,0,winW-dispcolWidth,150) ;
      addObject(hsvwheel, cast(HMENU)(CMD.HSV)) ;
      hsvwheel.setRGB(cur_color) ;
      
      rgbslider= new RGBslider("RGBslider","RGBslider") ;
      rgbslider.setStyle(WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE) ;
      rgbslider.setPosition(0,150,winW,84) ;
      rgbslider.enableAlpha(enbAlpha) ;
      addObject(rgbslider, cast(HMENU)(CMD.RGB)) ;
      rgbslider.setRGBA(cur_color) ;

      buttonOK= new PushButton("OK") ;
      buttonOK.addStyle(WS_TABSTOP) ;
      buttonOK.setPosition(winW-184,winH-32,80,20) ;
      addObject( buttonOK, cast(HMENU)(CMD.OK) ) ;
      buttonOK.setFont(hfont) ;

      buttonCAN= new PushButton("Cancel") ;
      buttonCAN.addStyle(WS_TABSTOP) ;
      buttonCAN.setPosition(winW-92,winH-32,80,20) ;
      addObject( buttonCAN, cast(HMENU)(CMD.CANCEL) ) ;
      buttonCAN.setFont(hfont) ;

      break ;
      
    case WM_PAINT:
      PAINTSTRUCT ps;
      HBRUSH DEFCOLB=CreateSolidBrush(def_color) ;
      HPEN   DEFCOLP=CreatePen(PS_SOLID,1,def_color) ;
      HBRUSH CURCOLB=CreateSolidBrush(cur_color) ;
      HPEN   CURCOLP=CreatePen(PS_SOLID,1,cur_color) ;
      
      HDC hdc = BeginPaint(hwnd, &ps);
      fillColor(false) ;
      StretchBlt(hdc, winW-dispcolWidth, 0, dispcolWidth, dispcolHeight,
		 WinDC_back, 0, 0, dispcolWidth, dispcolHeight, SRCCOPY);
      EndPaint(hwnd, &ps);
      DeleteObject(DEFCOLB) ;
      DeleteObject(DEFCOLP) ;
      DeleteObject(CURCOLB) ;
      DeleteObject(CURCOLP) ;
      break ;
      
    case WM_SIZE:
      break ;
      
    case WM_CLOSE:
      setResult(0) ;
      DestroyWindow(hwnd);
      break ;
      
    case WM_COMMAND:
      dlgstate=LOWORD(wp) ;
      switch( dlgstate ){
      case CMD.CANCEL:
        setResult(0) ;
        DestroyWindow(hwnd);
        break ;
      case CMD.OK:
        setResult(1) ;
        DestroyWindow(hwnd);
        break ;
      case CMD.HSV:
	rgbslider.setRGB(hsvwheel.getRGB()) ;
	cur_color = rgbslider.getRGBA() ;
	RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
	break ;
      case CMD.RGB:
	cur_color = rgbslider.getRGBA() ;
	hsvwheel.setRGB(cur_color) ;
	RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
	break ;
	
      default :
	break ;
      }
      break ;
      
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

  void enableAlpha(bool enbAlpha){
    this.enbAlpha=enbAlpha ;
  }
  
  int setColor(uint col){
    cur_color=col ;
    def_color=col ;
    return(0) ;
  }
  
  uint getColor(){
    return(cur_color) ;
  }

  int fillColor(bool init){
    if( init==true ){
      for( int y=0 ; y<dispcolHeight/2 ; y++ ){
	for( int x=0 ; x<dispcolWidth ; x++ ){
	  int bgv = ((x/8+y/8)%2)==0 ? 128 : 70 ;
	  int a = enbAlpha==true ? (def_color>>24) & 0x000000ff : 255 ;
	  int r = (def_color    ) & 0x000000ff ;
	  int g = (def_color>>8 ) & 0x000000ff ;
	  int b = (def_color>>16) & 0x000000ff ;
	  r = (r*a + bgv*(255-a)) /255 ;
	  g = (g*a + bgv*(255-a)) /255 ;
	  b = (b*a + bgv*(255-a)) /255 ;
	  bgmap[y*dispcolWidth+x]=RGBA(b,g,r,255) ;
	}
      }
    }
    for( int y=dispcolHeight/2 ; y<dispcolHeight ; y++ ){
      for( int x=0 ; x<dispcolWidth ; x++ ){
	  int bgv = ((x/8+y/8)%2)==0 ? 128 : 70 ;
	  int a = enbAlpha==true ? (cur_color>>24) & 0x000000ff : 255 ;
	  int r = (cur_color    ) & 0x000000ff ;
	  int g = (cur_color>>8 ) & 0x000000ff ;
	  int b = (cur_color>>16) & 0x000000ff ;
	  r = (r*a + bgv*(255-a)) /255 ;
	  g = (g*a + bgv*(255-a)) /255 ;
	  b = (b*a + bgv*(255-a)) /255 ;
	  bgmap[y*dispcolWidth+x]=RGBA(b,g,r,255) ;
      }
    }
    return(0) ;
  }
}
