import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.WindowObject;
import std.stdio;
import std.string;
import std.utf;

COLORREF RGBA(uint r, uint g, uint b, uint a) {
  return cast(COLORREF)(r | (g << 8) | (b << 16) | (a<<24));
}

const LBUTTONDOWN   = (0x0001) ;
const LBUTTONUP     = (0x0002) ;
const LBUTTONDBLCLK = (0x0004) ;
const RBUTTONDOWN   = (0x0010) ;
const RBUTTONUP     = (0x0020) ;
const RBUTTONDBLCLK = (0x0040) ;
const ALT_KEY       = (0x0100) ;
const CNTL_KEY      = (0x0200) ;
const SHIFT_KEY     = (0x0400) ;

class ImgBuf{
  uint     *pt    =null ;
  uint[][] ary ;
  int      dspW=0 ;
  int      dspH=0 ;
  int      orgW=0 ;
  int      orgH=0 ;
  int      sclW=0 ;
  int      sclH=0 ;
  double   scl =1.0 ;

  this(){}

  ~this(){
    freeArray() ;
  }

  void setImgBuf(int w, int h){
    orgW=w ;
    orgH=h ;
    freeArray() ;
    core.memory.GC.collect() ;
    core.memory.GC.minimize() ;
    ary.length = orgH ;
    for( int y=0 ; y<ary.length ; y++ ){
      ary[y] = new uint[orgW] ;
    }
    reScale() ;
  }

  private void freeArray(){
    for( int y=0 ; y<ary.length ; y++ ){
      delete ary[y] ;
    }
    delete ary ;
  }

  private void reScale(){
    sclW = cast(int)(cast(double)(orgW) * scl) ;
    sclH = cast(int)(cast(double)(orgH) * scl) ;
  }

  void setScale(double scl){
    this.scl=scl ;
    reScale() ;
  }

  double getScale(){
    return(scl) ;
  }
}

class GraphicBaseEx : WindowObject {
  HDC        WinDC;
  HDC        WinDC_back;
  HBITMAP    bmp_back  ;
  BITMAPINFO bi ;
  SCROLLINFO rollinfo_H ;
  SCROLLINFO rollinfo_V ;
  int        rollstep   ;
  bool       disnoscroll=false ;
  bool       centering  =false ;
  bool       autofocus  =false ;
  double     mouseX  =0 ;
  double     mouseY  =0 ;
  double     mouseWX =0 ;
  double     mouseWY =0 ;
  int        mouseB  =0 ;
  int        ACS_stat=0 ;
  ImgBuf     imgbuf ;

  this(string name, string ClassName) {
    winc.style          = 0;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.lpszMenuName   = null;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.lpszClassName  = toUTFz!(wchar*)(ClassName) ;
    super(name) ;
    this.registwinc() ;
    rollstep=20 ;
    imgbuf = new ImgBuf() ;
  }

  int setImageSize(int width, int height){
    if( width>0 && height>0 ){
      imgbuf.setImgBuf(width,height) ;
    }else{
      imgbuf.setImgBuf(0,0) ;
    }
    if( getHWND() != null ){
      reDisplay() ;
    }
    return(0) ;
  }

  int setScale(double scl){
    if( scl>0.0 ){
      imgbuf.setScale(scl) ;
      reDisplay() ;
      rollinfo_V.fMask  = SIF_POS ;
      rollinfo_V.nPos = cast(int)((mouseY*imgbuf.scl - mouseWY)) ;
      if( rollinfo_V.nPos < 0 ) rollinfo_V.nPos=0 ;
      if( rollinfo_V.nPos > (rollinfo_V.nMax-imgbuf.dspH) ) rollinfo_V.nPos=rollinfo_V.nMax-imgbuf.dspH ;
      SetScrollInfo(getHWND(), SB_VERT, &rollinfo_V, true) ;
      rollinfo_H.fMask  = SIF_POS ;
      rollinfo_H.nPos = cast(int)((mouseX*imgbuf.scl - mouseWX)) ;
      if( rollinfo_H.nPos < 0 ) rollinfo_H.nPos=0 ;
      if( rollinfo_H.nPos > (rollinfo_H.nMax-imgbuf.dspW) ) rollinfo_H.nPos=rollinfo_H.nMax-imgbuf.dspW ;
      SetScrollInfo(getHWND(), SB_HORZ, &rollinfo_H, true) ;
      InvalidateRect(getHWND() , null , true);
    }
    return(0) ;
  }

  double getScale(){
    return(imgbuf.getScale()) ;
  }

  int setRollstep(int step){
    rollstep=step ;
    return(0) ;
  }

  int reDisplay(){
    if( getHWND() != null ){
      sendmsg( getHWND(), WM_SIZE, 0, (winH<<16 | winW) );
      InvalidateRect(getHWND() , null , true);
    }
    return(0) ;
  }

  int setDisableNoScroll(bool flag){
    disnoscroll=flag ;
    return(0) ;
  }

  int setCentering(bool flag){
    centering = flag ;
    return(0) ;
  }
  
  int setAutoFocus(bool flag){
    autofocus = flag ;
    return(0) ;
  }
  
  override LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      this.wm_destroy() ;
      DeleteDC(WinDC);
      DeleteDC(WinDC_back);
      DeleteObject(bmp_back);
      break ;
      
    case WM_CREATE:
      imgbuf.dspW=winW ;
      imgbuf.dspH=winH ;
      
      bi.bmiHeader.biSize             = BITMAPINFOHEADER.sizeof;
      bi.bmiHeader.biWidth            =  imgbuf.dspW ;
      bi.bmiHeader.biHeight           = -imgbuf.dspH ;
      bi.bmiHeader.biPlanes           = 1;
      bi.bmiHeader.biBitCount         = 32;
      bi.bmiHeader.biCompression      = BI_RGB;
      bi.bmiHeader.biSizeImage        = 0;
      bi.bmiHeader.biXPelsPerMeter    = 0;
      bi.bmiHeader.biYPelsPerMeter    = 0;
      bi.bmiHeader.biClrUsed          = 1;
      bi.bmiHeader.biClrImportant     = 0;

      rollinfo_H.cbSize = SCROLLINFO.sizeof ;
      rollinfo_H.fMask  = SIF_RANGE | SIF_POS | SIF_PAGE | (disnoscroll==true ? SIF_DISABLENOSCROLL : 0) ;
      rollinfo_H.nPos   = 0         ;
      rollinfo_H.nPage  = imgbuf.dspW-1 ;
      rollinfo_H.nMin   = 0         ;
      rollinfo_H.nMax   = imgbuf.sclW-1 ;
      SetScrollInfo(hwnd, SB_HORZ, &rollinfo_H, true) ;
      
      rollinfo_V.cbSize = SCROLLINFO.sizeof ;
      rollinfo_V.fMask  = SIF_RANGE | SIF_POS | SIF_PAGE | (disnoscroll==true ? SIF_DISABLENOSCROLL : 0) ;
      rollinfo_V.nPos   = 0          ;
      rollinfo_V.nPage  = imgbuf.dspH-1 ;
      rollinfo_V.nMin   = 0          ;
      rollinfo_V.nMax   = imgbuf.sclH-1 ;
      SetScrollInfo(hwnd, SB_VERT, &rollinfo_V, true) ;

      RECT wrect ;
      GetClientRect( hwnd, &wrect ) ;

      rollinfo_H.fMask  = SIF_RANGE | SIF_PAGE | (disnoscroll==true ? SIF_DISABLENOSCROLL : 0) ;
      rollinfo_H.nPage  = wrect.right;
      rollinfo_H.nMin   = 0          ;
      rollinfo_H.nMax   = imgbuf.sclW-1 ;
      SetScrollInfo(hwnd, SB_HORZ, &rollinfo_H, true) ;
      rollinfo_H.fMask  = SIF_POS ;
      
      rollinfo_V.fMask  = SIF_RANGE | SIF_PAGE | (disnoscroll==true ? SIF_DISABLENOSCROLL : 0) ;
      rollinfo_V.nPage  = wrect.bottom;
      rollinfo_V.nMin   = 0           ;
      rollinfo_V.nMax   = imgbuf.sclH-1 ;
      SetScrollInfo(hwnd, SB_VERT, &rollinfo_V, true) ;
      rollinfo_V.fMask  = SIF_POS ;
      
      WinDC=GetDC(hwnd);
      WinDC_back=CreateCompatibleDC(WinDC);
      bmp_back  =CreateDIBSection( WinDC, &bi, DIB_RGB_COLORS, cast(void**)(&imgbuf.pt), null, 0 );
      SelectObject(WinDC_back, bmp_back);
      
      this.wm_create() ;
      break ;

    case WM_HSCROLL:
      {
	bool changed=true ;
	
	switch( LOWORD(wp) ){
	case SB_LINELEFT :
	  rollinfo_H.nPos -=rollstep ;
	  if( rollinfo_H.nPos < 0 ){
	    rollinfo_H.nPos=0 ;
	  }
	  break ;
	case SB_LINERIGHT :
	  rollinfo_H.nPos +=rollstep ;
	  if( rollinfo_H.nPos >= (rollinfo_H.nMax-imgbuf.dspW) ){
	    rollinfo_H.nPos=rollinfo_H.nMax-imgbuf.dspW+1 ;
	  }
	  break ;
	case SB_PAGELEFT :
	  rollinfo_H.nPos -= rollinfo_H.nPage/2 ;
	  if( rollinfo_H.nPos < 0 ){
	    rollinfo_H.nPos=0 ;
	  }
	  break ;
	case SB_PAGERIGHT :
	  rollinfo_H.nPos += rollinfo_H.nPage/2 ;
	  if( rollinfo_H.nPos >= (rollinfo_H.nMax-imgbuf.dspW) ){
	    rollinfo_H.nPos=rollinfo_H.nMax-imgbuf.dspW+1 ;
	  }
	  break ;
	case SB_THUMBTRACK :
	  {
	    SCROLLINFO  tmp_scrinfo ;
	    tmp_scrinfo.cbSize = SCROLLINFO.sizeof ; tmp_scrinfo.fMask = SIF_TRACKPOS ;
	    GetScrollInfo( hwnd, SB_HORZ, &tmp_scrinfo) ;
	    rollinfo_H.nPos = tmp_scrinfo.nTrackPos ;
	  }
	  break ;
	  
	default:
	  changed=false ;
	  break ;
	}
      
	if( changed==true ){
	  SetScrollInfo( hwnd, SB_HORZ, &rollinfo_H, true) ;
	  setCurPosition(winW/2, winH/2) ;
	  reDisplay() ;
	}
      }
      break ;
      
    case WM_VSCROLL:
      {
	bool changed=true ;
	
	switch( LOWORD(wp) ){
	case SB_LINEUP :
	  rollinfo_V.nPos -=rollstep ;
	  if( rollinfo_V.nPos < 0 ){
	    rollinfo_V.nPos=0 ;
	  }
	  break ;
	case SB_LINEDOWN :
	  rollinfo_V.nPos +=rollstep ;
	  if( rollinfo_V.nPos >= (rollinfo_V.nMax-imgbuf.dspH) ){
	    rollinfo_V.nPos=rollinfo_V.nMax-imgbuf.dspH+1 ;
	  }
	  break ;
	case SB_PAGEUP :
	  rollinfo_V.nPos -= rollinfo_V.nPage/2 ;
	  if( rollinfo_V.nPos < 0 ){
	    rollinfo_V.nPos=0 ;
	  }
	  break ;
	case SB_PAGEDOWN :
	  rollinfo_V.nPos += rollinfo_V.nPage/2 ;
	  if( rollinfo_V.nPos >= (rollinfo_V.nMax-imgbuf.dspH) ){
	    rollinfo_V.nPos=rollinfo_V.nMax-imgbuf.dspH+1 ;
	  }
	  break ;
	case SB_THUMBTRACK :
	  {
	    SCROLLINFO  tmp_scrinfo ;
	    tmp_scrinfo.cbSize = SCROLLINFO.sizeof ; tmp_scrinfo.fMask = SIF_TRACKPOS ;
	    GetScrollInfo( hwnd, SB_VERT, &tmp_scrinfo) ;
	    rollinfo_V.nPos = tmp_scrinfo.nTrackPos ;
	  }
	  break ;
	  
	default:
	  changed=false ;
	  break ;
	}
	if( changed==true ){
	  SetScrollInfo( hwnd, SB_VERT, &rollinfo_V, true) ;
	  setCurPosition(winW/2, winH/2) ;
	  reDisplay() ;
	}
      }
      break ;
      
    case WM_PAINT:
      PAINTSTRUCT ps;
      HDC         hdc;
      SCROLLINFO  slinfo_H ;
      SCROLLINFO  slinfo_V ;
      HBRUSH      COLBG  ;
      HPEN        PCOLBG ;

      this.wm_paint() ;
      
      slinfo_H.cbSize = SCROLLINFO.sizeof ; slinfo_H.fMask = SIF_POS ;
      slinfo_V.cbSize = SCROLLINFO.sizeof ; slinfo_V.fMask = SIF_POS ;
      GetScrollInfo( hwnd, SB_HORZ, &slinfo_H) ;
      GetScrollInfo( hwnd, SB_VERT, &slinfo_V) ;

      PCOLBG = CreatePen(PS_SOLID, 1, RGBA(128,128,128,0) & 0x00ffffff) ;
      COLBG  = CreateSolidBrush(RGBA(128,128,128,0) & 0x00ffffff) ;
      hdc = BeginPaint(hwnd, &ps);
      SelectObject(hdc, PCOLBG) ;
      SelectObject(hdc, COLBG) ;
      if( imgbuf.sclW<=0 || imgbuf.sclH<=0 ){
	Rectangle(hdc, 0,0,imgbuf.dspW,imgbuf.dspH) ;
      }else{
	int coffx    = (imgbuf.dspW-imgbuf.sclW)/2 ;
	int coffy    = (imgbuf.dspH-imgbuf.sclH)/2 ;
	int dst_posx = centering==true ? (imgbuf.sclW<=imgbuf.dspW ? coffx : 0) : 0 ;
	int dst_posy = centering==true ? (imgbuf.sclH<=imgbuf.dspH ? coffy : 0) : 0 ;
	int src_posx = slinfo_H.nPos ;
	int src_posy = slinfo_V.nPos ;
	int width    = imgbuf.sclW<=imgbuf.dspW ? imgbuf.sclW : imgbuf.dspW ;
	int height   = imgbuf.sclH<=imgbuf.dspH ? imgbuf.sclH : imgbuf.dspH ;

	for( int y=0 ; y<height ; y++ ){
	  int dstx = dst_posx ;
	  int dsty = dst_posy+y ;
	  int srcy = cast(int)(cast(double)(y+src_posy)/imgbuf.scl);
	  srcy = srcy<0 ? 0 : srcy>imgbuf.orgH ? imgbuf.orgH-1 : srcy ;
	  for( int x=0 ; x<width ; x++ ){
	    int srcx = cast(int)(cast(double)(x+src_posx)/imgbuf.scl);
	    srcx = srcx<0 ? 0 : srcx>imgbuf.orgW ? imgbuf.orgW-1 : srcx ;
	    imgbuf.pt[(dsty*winW+dstx+x)] = imgbuf.ary[srcy][srcx] ;
	  }
	}
	
	SelectObject(WinDC_back, PCOLBG) ;
	SelectObject(WinDC_back, COLBG) ;
	if( centering==false ){
	  if( imgbuf.dspW > imgbuf.sclW ){
	    Rectangle(WinDC_back, imgbuf.sclW,0,imgbuf.dspW,imgbuf.dspH) ;
	  }
	  if( imgbuf.dspH > imgbuf.sclH ){
	    Rectangle(WinDC_back, 0,imgbuf.sclH,imgbuf.dspW,imgbuf.dspH) ;
	  }
	}else{
	  if( imgbuf.dspW > imgbuf.sclW ){
	    Rectangle(WinDC_back,                 0,0, coffx      ,imgbuf.dspH) ;
	    Rectangle(WinDC_back, imgbuf.sclW+coffx,0, imgbuf.dspW,imgbuf.dspH) ;
	  }
	  if( imgbuf.dspH > imgbuf.sclH ){
	    Rectangle(WinDC_back, 0,                0, imgbuf.dspW,coffy      ) ;
	    Rectangle(WinDC_back, 0,imgbuf.sclH+coffy, imgbuf.dspW,imgbuf.dspH) ;
	  }
	}
	StretchBlt(hdc,        0, 0, imgbuf.dspW, imgbuf.dspH,
		   WinDC_back, 0, 0, imgbuf.dspW, imgbuf.dspH, SRCCOPY);
      }
      EndPaint(hwnd, &ps);
      DeleteObject(PCOLBG) ;
      DeleteObject(COLBG) ;
      break ;

    case WM_WINDOWPOSCHANGING :
      rollinfo_H.fMask  = SIF_RANGE ;
      rollinfo_H.nMin   = 0         ;
      rollinfo_H.nMax   = imgbuf.sclW-1 ;
      SetScrollInfo(hwnd, SB_HORZ, &rollinfo_H, true) ;
      rollinfo_H.fMask  = SIF_POS ;
      
      rollinfo_V.fMask  = SIF_RANGE ;
      rollinfo_V.nMin   = 0          ;
      rollinfo_V.nMax   = imgbuf.sclH-1 ;
      SetScrollInfo(hwnd, SB_VERT, &rollinfo_V, true) ;
      rollinfo_V.fMask  = SIF_POS ;
      break ;
      
    case WM_SIZE:
      imgbuf.dspW=LOWORD(lp) ;
      imgbuf.dspH=HIWORD(lp) ;

      DeleteDC(WinDC_back);
      DeleteObject(bmp_back);
      bi.bmiHeader.biWidth  =  imgbuf.dspW ;
      bi.bmiHeader.biHeight = -imgbuf.dspH ;
      WinDC_back=CreateCompatibleDC(WinDC);
      bmp_back  =CreateDIBSection( WinDC, &bi, DIB_RGB_COLORS, cast(void**)(&imgbuf.pt), null, 0 );
      SelectObject(WinDC_back, bmp_back);

      rollinfo_H.fMask  = SIF_RANGE | SIF_PAGE | (disnoscroll==true ? SIF_DISABLENOSCROLL : 0) ;
      rollinfo_H.nPage  = imgbuf.dspW ;
      rollinfo_H.nMin   = 0         ;
      rollinfo_H.nMax   = imgbuf.sclW-1 ;
      SetScrollInfo(hwnd, SB_HORZ, &rollinfo_H, true) ;
      rollinfo_H.fMask  = SIF_POS ;
      
      rollinfo_V.fMask  = SIF_RANGE | SIF_PAGE | (disnoscroll==true ? SIF_DISABLENOSCROLL : 0) ;
      rollinfo_V.nPage  = imgbuf.dspH ;
      rollinfo_V.nMin   = 0          ;
      rollinfo_V.nMax   = imgbuf.sclH-1 ;
      SetScrollInfo(hwnd, SB_VERT, &rollinfo_V, true) ;
      rollinfo_V.fMask  = SIF_POS ;

      RECT wrect ;
      GetClientRect( hwnd, &wrect ) ;

      rollinfo_H.fMask  = SIF_RANGE | SIF_PAGE | (disnoscroll==true ? SIF_DISABLENOSCROLL : 0) ;
      rollinfo_H.nPage  = wrect.right;
      rollinfo_H.nMin   = 0          ;
      rollinfo_H.nMax   = imgbuf.sclW-1 ;
      SetScrollInfo(hwnd, SB_HORZ, &rollinfo_H, true) ;
      rollinfo_H.fMask  = SIF_POS ;
      
      rollinfo_V.fMask  = SIF_RANGE | SIF_PAGE | (disnoscroll==true ? SIF_DISABLENOSCROLL : 0) ;
      rollinfo_V.nPage  = wrect.bottom;
      rollinfo_V.nMin   = 0           ;
      rollinfo_V.nMax   = imgbuf.sclH-1 ;
      SetScrollInfo(hwnd, SB_VERT, &rollinfo_V, true) ;
      rollinfo_V.fMask  = SIF_POS ;
      
      this.wm_size(imgbuf.dspW,imgbuf.dspH) ;

      break;

    case WM_MOUSEWHEEL:
      int dlt=cast(int)(wp)>>16 ;
      int key=LOWORD(wp) ;
      int ret=wm_mousewheel(dlt,key) ;
      if( ret==0 ){
	if( (key&MK_SHIFT)==0 ){
	  rollinfo_V.fMask  = SIF_POS ;
	  rollinfo_V.nPos -=dlt/4 ;
	  if( rollinfo_V.nPos < 0 ) rollinfo_V.nPos=0 ;
	  if( rollinfo_V.nPos >= (rollinfo_V.nMax-imgbuf.dspH) ) rollinfo_V.nPos=rollinfo_V.nMax-imgbuf.dspH+1 ;
	  SetScrollInfo(hwnd, SB_VERT, &rollinfo_V, true) ;
	  setCurPosition(0,0,true);
	  reDisplay() ;
	}else{
	  rollinfo_H.fMask  = SIF_POS ;
	  rollinfo_H.nPos -=dlt/4 ;
	  if( rollinfo_H.nPos < 0 ) rollinfo_H.nPos=0 ;
	  if( rollinfo_H.nPos >= (rollinfo_H.nMax-imgbuf.dspW) ) rollinfo_H.nPos=rollinfo_H.nMax-imgbuf.dspW+1 ;
	  SetScrollInfo(hwnd, SB_HORZ, &rollinfo_H, true) ;
	  setCurPosition(0,0,true);
	  reDisplay() ;
	}
      }
      break ;

    case WM_LBUTTONDOWN:
    case WM_LBUTTONUP:
    case WM_LBUTTONDBLCLK:
    case WM_RBUTTONDOWN:
    case WM_RBUTTONUP:
    case WM_RBUTTONDBLCLK:
    case WM_MOUSEMOVE:
      {
	POINT point;

	if( msg==WM_LBUTTONDOWN || msg==WM_RBUTTONDOWN ) SetCapture( hwnd ) ;
	if( msg==WM_LBUTTONUP   || msg==WM_RBUTTONUP   ) ReleaseCapture() ;
	if( msg==WM_MOUSEMOVE ){
	  TRACKMOUSEEVENT trackmsevent ;
	  trackmsevent.cbSize  = TRACKMOUSEEVENT.sizeof ;
	  trackmsevent.dwFlags = TME_LEAVE ;
	  trackmsevent.hwndTrack = hwnd ;
	  TrackMouseEvent( &trackmsevent ) ;
	}

	GetCursorPos(&point);
	ScreenToClient( hwnd, &point );

	if( autofocus==true ) SetFocus(hwnd) ;
	setCurPosition(point.x, point.y) ;

	if(GetAsyncKeyState(VK_LBUTTON) < 0) mouseB |=   LBUTTONDOWN  ;
	else                                 mouseB &= ~(LBUTTONDOWN) ;
	if(GetAsyncKeyState(VK_RBUTTON) < 0) mouseB |=   RBUTTONDOWN  ;
	else                                 mouseB &= ~(RBUTTONDOWN) ;

	if(GetKeyState(VK_SHIFT)   < 0) ACS_stat |=   SHIFT_KEY  ;
	else                            ACS_stat &= ~(SHIFT_KEY) ;
	if(GetKeyState(VK_CONTROL) < 0) ACS_stat |=   CNTL_KEY   ;
	else                            ACS_stat &= ~(CNTL_KEY)  ;
	if(GetKeyState(VK_MENU)    < 0) ACS_stat |=   ALT_KEY    ;
	else                            ACS_stat &= ~(ALT_KEY)   ;
	if(GetKeyState(VK_SPACE)   < 0){
	  ACS_stat |=   ALT_KEY    ;
	  ACS_stat |=   SHIFT_KEY  ;
	}

	if( this.mouse(ACS_stat, mouseB, cast(int)(mouseX), cast(int)(mouseY))!=0 ){
	  reDisplay() ;
	}
      }
      break ;

    case WM_MOUSELEAVE:
      setCurPosition(winW/2, winH/2) ;
      reDisplay() ;
      break ;

    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

  void setCurPosition(double x, double y, bool recalc=false){
    SCROLLINFO slinfo_H ;
    SCROLLINFO slinfo_V ;

    slinfo_H.cbSize = SCROLLINFO.sizeof ; slinfo_H.fMask = SIF_TRACKPOS ;
    slinfo_V.cbSize = SCROLLINFO.sizeof ; slinfo_V.fMask = SIF_TRACKPOS ;
    GetScrollInfo( getHWND(), SB_HORZ, &slinfo_H) ;
    GetScrollInfo( getHWND(), SB_VERT, &slinfo_V) ;

    if( recalc==false ){
      mouseWX = x ;
      mouseWY = y ;
    }
    mouseX = cast(double)(mouseWX + slinfo_H.nTrackPos - (centering==true ? (imgbuf.sclW<=imgbuf.dspW ? (imgbuf.dspW-imgbuf.sclW)/2 : 0) : 0))/imgbuf.scl ;
    mouseY = cast(double)(mouseWY + slinfo_V.nTrackPos - (centering==true ? (imgbuf.sclH<=imgbuf.dspH ? (imgbuf.dspH-imgbuf.sclH)/2 : 0) : 0))/imgbuf.scl ;
  }

  int wm_destroy(){
    return(0) ;
  }
  int wm_create(){
    return(0) ;
  }
  int wm_paint(){
    return(0) ;
  }
  int wm_size(int w, int h){
    return(0) ;
  }
  int wm_mousewheel(int dlt, int key){
    return(0) ;
  }

  uint[][] getImageBuffer(){
    return(imgbuf.ary) ;
  }
  int getImageBufferWidth(){
    return(imgbuf.orgW) ;
  }
  int getImageBufferHeight(){
    return(imgbuf.orgH) ;
  }

  int mouse( int acs_stat, int mouseb, int mousex, int mousey ){
    return(0) ;
  }

}
