import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.WindowObject;
import std.stdio;
import std.string;
import std.utf;
import std.math;

T ABS(T)(T a) { return( a<0 ? -a : a ) ;}

COLORREF RGBA(uint r, uint g, uint b, uint a) {
  return cast(COLORREF)(r | (g << 8) | (b << 16) | (a<<24));
}

const LBUTTONDOWN   = (0x0001) ;
const LBUTTONUP     = (0x0002) ;
const LBUTTONDBLCLK = (0x0004) ;
const RBUTTONDOWN   = (0x0010) ;
const RBUTTONUP     = (0x0020) ;
const RBUTTONDBLCLK = (0x0040) ;
const ALT_KEY       = (0x0100) ;
const CNTL_KEY      = (0x0200) ;
const SHIFT_KEY     = (0x0400) ;

struct vec2D  {
  double  x ;
  double  y ;

  void opAssign(double x, double y){
    this.x = x ;
    this.y = y ;
  }

  void opAssign(vec2D d){
    this.x = d.x ;
    this.y = d.y ;
  }

  vec2D opUnary(string s)(){
    static if( s=="-" ){
      vec2D r ;
      r.x = -this.x ;
      r.y = -this.y ;
      return(r) ;
    }
  }

  vec2D opBinary(string op)(vec2D d){
    static if( op=="+" ){
      vec2D r ;
      r.x = this.x + d.x ;
      r.y = this.y + d.y ;
      return(r) ;
    } else static if( op=="-" ){
      vec2D r ;
      r.x = this.x - d.x ;
      r.y = this.y - d.y ;
      return(r) ;
    } else static assert(0,"Operator " ~ op ~ " does not exist") ;
  }

  vec2D opBinary(string op)(double d){
    static if( op=="+" ){
      vec2D r ;
      r.x = this.x + d ;
      r.y = this.y + d ;
      return(r) ;
    } else static if( op=="-" ){
      vec2D r ;
      r.x = this.x - d ;
      r.y = this.y - d ;
      return(r) ;
    } else static if( op=="*" ){
      vec2D r ;
      r.x = this.x * d ;
      r.y = this.y * d ;
      return(r) ;
    } else static if( op=="/" ){
      vec2D r ;
      r.x = this.x / d ;
      r.y = this.y / d ;
      return(r) ;
    } else static assert(0,"Operator " ~ op ~ " does not exist") ;
  }

  void opOpAssign(string op)(vec2D d){
    static if( op=="+" ){
      this.x += d.x ;
      this.y += d.y ;
    } else static if( op=="-" ){
      this.x -= d.x ;
      this.y -= d.y ;
    } else static assert(0,"Operator " ~ op ~ " does not exist") ;
  }

  void opOpAssign(string op)(double d){
    static if( op=="+" ){
      this.x += d ;
      this.y += d ;
    } else static if( op=="-" ){
      this.x -= d ;
      this.y -= d ;
    } else static if( op=="*" ){
      this.x *= d ;
      this.y *= d ;
    } else static if( op=="/" ){
      this.x /= d ;
      this.y /= d ;
    } else static assert(0,"Operator " ~ op ~ " does not exist") ;
  }

  void inv(){
    this.x = -this.x ;
    this.y = -this.y ;
  }

  const bool opEquals( ref const(vec2D) b ){
    if( this.x == b.x &&
	this.y == b.y ) {
      return(true) ;
    }
    return(false) ;
  }

  double distance( vec2D b ){
    double lx=this.x-b.x ;
    double ly=this.y-b.y ;
    return( sqrt( lx*lx + ly*ly ) ) ;
  }

  double length(){
    return( sqrt( this.x*this.x + this.y*this.y ) ) ;
  }

  vec2D normalize(){
    vec2D  r ;
    double d ;
    d=sqrt( this.x*this.x + this.y*this.y ) ;
    if( d==0.0 ){
      r.x=0.0 ;
      r.y=0.0 ;
    }else{
      r.x = this.x/d ;
      r.y = this.y/d ;
    }
    return(r) ;
  }

  double dot( vec2D b ){
    return( (this.x * b.x) + (this.y * b.y) ) ;
  }

  double cross( vec2D b ){
    return((this.x * b.y) - (this.y * b.x)) ;
  }

  vec2D rotate( double angle ){
    vec2D r ;
    double sinval = sin(angle);
    double cosval = cos(angle);
    r.x = cosval*this.x - sinval*this.y ;
    r.y = sinval*this.x + cosval*this.y ;
    return(r) ;
  }

  void print() {
    writef("x=%f, y=%f\n",x,y) ;
  }
  
} ;

class HSVwheel : WindowObject {
  int mouseX  =0 ;
  int mouseY  =0 ;
  int mouseB  =0 ;
  int ACS_stat=0 ;
  int touching   ;
  int cur_mode=0 ;
  double cur_hue ; //-pi〜pi
  double cur_val ; //0〜1.0
  double cur_sat ; //0〜1.0
  double cur_cposx  ;
  double cur_cposy  ;
  double cur_cposx0 ;
  double cur_cposy0 ;
  double cur_cposx1 ;
  double cur_cposy1 ;
  double cur_tposx  ;
  double cur_tposy  ;

  HDC        WinDC;
  HDC        WinDC_back;
  HBITMAP    bmp_back  ;
  BITMAPINFO bi ;
  uint*      bgmap ;
  uint[]     bgcircle ;
  uint[]     bgcletri ;
  uint       bg_color ;
  int        sample=4 ;
  double     circlew  ;
  double     trir     ;

  this(string name, string ClassName) {
    winc.style          = 0;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.lpszMenuName   = null;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.lpszClassName  = toUTFz!(wchar*)(ClassName) ;
    super(name) ;
    this.registwinc() ;
    circlew=16.0 ;
    trir=circlew+2.0 ;
    bg_color = RGBA(192,192,192,255) ;
  }

  void setBGCOLOR(int r, int g, int b){
    bg_color = RGBA(r,g,b,255) ;
  }

  override LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      DeleteObject(bmp_back);
      DeleteDC(WinDC_back);
      ReleaseDC(hwnd,WinDC);
      break ;
      
    case WM_CREATE:
      bi.bmiHeader.biSize             = BITMAPINFOHEADER.sizeof;
      bi.bmiHeader.biWidth            =  (winW>0) ?  winW :  2 ;
      bi.bmiHeader.biHeight           =  (winH>0) ? -winH : -2 ;
      bi.bmiHeader.biPlanes           = 1;
      bi.bmiHeader.biBitCount         = 32;
      bi.bmiHeader.biCompression      = BI_RGB;
      bi.bmiHeader.biSizeImage        = 0;
      bi.bmiHeader.biXPelsPerMeter    = 0;
      bi.bmiHeader.biYPelsPerMeter    = 0;
      bi.bmiHeader.biClrUsed          = 1;
      bi.bmiHeader.biClrImportant     = 0;
      
      WinDC=GetDC(hwnd);
      WinDC_back=CreateCompatibleDC(WinDC);
      bmp_back  =CreateDIBSection( WinDC, &bi, DIB_RGB_COLORS, cast(void**)(&bgmap), null, 0 );
      SelectObject(WinDC_back, bmp_back);
      bgcircle.length = bi.bmiHeader.biWidth * -bi.bmiHeader.biHeight ;
      bgcletri.length = bi.bmiHeader.biWidth * -bi.bmiHeader.biHeight ;
      
      cur_hue=0.0 ;
      cur_val=0.0 ;
      cur_sat=0.0 ;
      break ;
      
    case WM_PAINT:
      HDC         hdc;
      PAINTSTRUCT ps ;
      HBRUSH      brush ;
      HPEN        pen ;
      
      HSVtoCURPOS(cur_hue, cur_sat, cur_val) ;
      drawColorTriangle(cur_hue, cur_sat, cur_val) ;
      foreach( i, p ; bgcletri)	bgmap[i]=p ;
      drawLine(cur_cposx0, cur_cposy0, cur_cposx1, cur_cposy1, cur_hue>=0.0 ? RGBA(0,0,0,0) : RGBA(255,255,255,0)) ;
      drawCircle(cur_tposx, cur_tposy, 4, cur_val<0.5 ? RGBA(0,0,0,0) : RGBA(255,255,255,0)) ;
      
      brush = CreateSolidBrush(RGBA(0,0,0,0)) ;
      pen   = CreatePen( PS_SOLID, 2, RGBA(255,255,255,0) ) ;
      hdc   = BeginPaint(hwnd, &ps);
      
      StretchBlt(hdc, 0, 0, winW, winH, WinDC_back, 0, 0, winW, winH, SRCCOPY);
      SelectObject(hdc, pen) ;
      SetROP2(hdc , R2_XORPEN);
      SelectObject(hdc, brush) ;
      
      EndPaint(hwnd, &ps);
      DeleteObject(pen) ;
      DeleteObject(brush) ;
      
      //printStatusbar( format("h=%f s=%f v=%f",cur_hue, cur_sat, cur_val) ) ;
      if( touching==1 ){
	sendmsgtoparent( WM_COMMAND, cast(WPARAM)(getHMENU()), 0) ;
      }
      break ;

    case WM_MOVE:
      RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
      break ;

    case WM_SIZE:
      if( bi.bmiHeader.biWidth!=winW || -bi.bmiHeader.biHeight!=winH || bmp_back != null ){
	DeleteDC(WinDC_back);
	DeleteObject(bmp_back);
	bi.bmiHeader.biWidth            =  (winW>0) ?  winW :  2 ;
	bi.bmiHeader.biHeight           =  (winH>0) ? -winH : -2 ;
	WinDC_back=CreateCompatibleDC(WinDC);
	bmp_back  =CreateDIBSection( WinDC, &bi, DIB_RGB_COLORS, cast(void**)(&bgmap), null, 0 );
	SelectObject(WinDC_back, bmp_back);
	bgcircle.length = bi.bmiHeader.biWidth * -bi.bmiHeader.biHeight ;
	bgcletri.length = bi.bmiHeader.biWidth * -bi.bmiHeader.biHeight ;
	int winsize = winW<winH ? winW/2 : winH/2 ;
	if( winsize<80 ){
	  circlew=2+14*winsize/80 ;
	  trir=circlew+2.0 ;
	}else{
	  circlew=16.0 ;
	  trir=circlew+2.0 ;
	}
	fillWheel() ;
      }
      RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
      break;

    case WM_LBUTTONDOWN:
    case WM_LBUTTONUP:
    case WM_LBUTTONDBLCLK:
    case WM_RBUTTONDOWN:
    case WM_RBUTTONUP:
    case WM_RBUTTONDBLCLK:
    case WM_MOUSEMOVE:
      {
	POINT point;

	GetCursorPos(&point);
	ScreenToClient( hwnd, &point );
	mouseX = point.x ;
	mouseY = point.y ;
	if( msg==WM_LBUTTONDOWN ){
	  cur_mode = checkSelectMode(mouseX, mouseY)==1 ? 2 : 1 ;
	  touching=1 ;
	  SetCapture( hwnd ) ;
	}
	if( msg==WM_LBUTTONUP   ){
	  touching=0 ;
	  cur_mode=0 ;
	  ReleaseCapture() ;
	}
	if( msg==WM_MOUSEMOVE ){
	}else{
	  mouseB=0 ;
	  switch( msg ){
	  case WM_LBUTTONDOWN:
	    mouseB|=LBUTTONDOWN ;
	    break ;
	  case WM_LBUTTONUP:
	    mouseB|=LBUTTONUP ;
	    break ;
	  case WM_LBUTTONDBLCLK:
	    mouseB|=LBUTTONDBLCLK ;
	    break ;
	  case WM_RBUTTONDOWN:
	    mouseB|=RBUTTONDOWN ;
	    break ;
	  case WM_RBUTTONUP:
	    mouseB|=RBUTTONUP ;
	    break ;
	  case WM_RBUTTONDBLCLK:
	    mouseB|=RBUTTONDBLCLK ;
	    break ;
	  default : break ;
	  }	  
	  
	  if(GetKeyState(VK_SHIFT)   < 0) ACS_stat |=   SHIFT_KEY  ;
	  else                            ACS_stat &= ~(SHIFT_KEY) ;
	  if(GetKeyState(VK_CONTROL) < 0) ACS_stat |=   CNTL_KEY   ;
	  else                            ACS_stat &= ~(CNTL_KEY)  ;
	  if(GetKeyState(VK_MENU)    < 0) ACS_stat |=   ALT_KEY    ;
	  else                            ACS_stat &= ~(ALT_KEY)   ;
	  
	}
	if( (mouseB & LBUTTONDOWN)!=0 ){
	  POStoHSV(mouseX, mouseY) ;
	  RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
	}
      }
      break ;
      
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }


  int HSVtoCURPOS(double hue, double sat, double val){

    double ro ;
    if( winW<winH ) ro = cast(double)(winW)/2.0 ;
    else            ro = cast(double)(winH)/2.0 ;
    cur_cposx  = (ro-circlew/2.0)*cos(-hue) + cast(double)(winW)/2.0 ;
    cur_cposy  = (ro-circlew/2.0)*sin(-hue) + cast(double)(winH)/2.0 ;
    cur_cposx0 = (ro-circlew+1)*cos(-hue) + cast(double)(winW)/2.0 ;
    cur_cposy0 = (ro-circlew+1)*sin(-hue) + cast(double)(winH)/2.0 ;
    cur_cposx1 = (ro        -1)*cos(-hue) + cast(double)(winW)/2.0 ;
    cur_cposy1 = (ro        -1)*sin(-hue) + cast(double)(winH)/2.0 ;

    vec2D  c = {(ro-trir)*cos( 0.0        ), (ro-trir)*sin( 0.0        )} ;
    vec2D  v = {(ro-trir)*cos(-PI*2.0/3.0 ), (ro-trir)*sin(-PI*2.0/3.0 )} ;
    vec2D  s = {(ro-trir)*cos( PI*2.0/3.0 ), (ro-trir)*sin( PI*2.0/3.0 )} ;

    vec2D dv = {1.0,0.0} ; dv=(dv.rotate(-PI*2.0/3.0))*((ro-trir)*3.0/2.0*val) ;
    vec2D ds = {1.0,0.0} ; ds=(ds.rotate( PI*2.0/3.0))*((ro-trir)*3.0/2.0*sat) ;

    vec2D v1=s+dv ;
    vec2D v2=c+dv ;
    vec2D s1=v+ds ;
    vec2D s2=c+ds ;
    
    vec2D lv=v2-v1 ;
    vec2D ls=s2-s1 ;

    vec2D v1s1=s1-v1 ;

    lv*=(ls.cross(v1s1)/ls.cross(lv)) ;
    v1+=lv ;
    v1=v1.rotate(-hue) ;

    cur_tposx = v1.x + (winW/2.0) ;
    cur_tposy = v1.y + (winH/2.0) ;
    
    return(0) ;
  }
  

  int checkSelectMode(double x, double y){
    vec2D  p  = {x - cast(double)(winW)/2.0, y - cast(double)(winH)/2.0} ;
    double rp = p.length() ;
    double th = acos( p.x/rp ) ;
    if( p.y<0 ) th=-th ;
    
    double ro, ri ;
    if( winW<winH ) ro = cast(double)(winW)/2.0 ;
    else            ro = cast(double)(winH)/2.0 ;
    ri = ro - trir ;

    th=-cur_hue ;
    vec2D  c = {ri*cos(th              ), ri*sin(th              )} ;
    vec2D  v = {ri*cos(th + PI*2.0/3.0 ), ri*sin(th + PI*2.0/3.0 )} ;
    vec2D  s = {ri*cos(th - PI*2.0/3.0 ), ri*sin(th - PI*2.0/3.0 )} ;
    
    vec2D  vecPC=c-p ;
    vec2D  vecPV=v-p ;
    vec2D  vecPS=s-p ;
    
    double crsCV = vecPC.cross(vecPV) ;
    double crsVS = vecPV.cross(vecPS) ;
    double crsSC = vecPS.cross(vecPC) ;
    
    if( crsCV>=0 && crsVS>=0 && crsSC>=0 ){ //check inside
      return(1) ;
    }
    return(0) ;
  }

  
  int POStoHSV(double x, double y){
    vec2D  p  = {x - cast(double)(winW)/2.0, y - cast(double)(winH)/2.0} ;
    double rp = p.length() ;
    double th = acos( p.x/rp ) ;
    if( p.y>=0 ) th=-th ;
    
    double ro, ri ;
    if( winW<winH ) ro = cast(double)(winW)/2.0 ;
    else            ro = cast(double)(winH)/2.0 ;
    ri = ro - trir ;

    if( cur_mode==1 ){
      cur_hue = th ;
    }else if( cur_mode==2 ){
      th=-cur_hue ;
      vec2D  c = {ri*cos(th              ), ri*sin(th              )} ;
      vec2D  v = {ri*cos(th + PI*2.0/3.0 ), ri*sin(th + PI*2.0/3.0 )} ;
      vec2D  s = {ri*cos(th - PI*2.0/3.0 ), ri*sin(th - PI*2.0/3.0 )} ;

      vec2D  vecPC=c-p ;
      vec2D  vecPV=v-p ;
      vec2D  vecPS=s-p ;

      double crsCV = vecPC.cross(vecPV) ;
      double crsVS = vecPV.cross(vecPS) ;
      double crsSC = vecPS.cross(vecPC) ;
	
      if( crsCV>=0 && crsVS>=0 && crsSC>=0 ){ //check inside
	double sumS = crsCV+crsVS+crsSC ;
	cur_sat = crsSC/(sumS) ;
	cur_val = crsCV/(sumS) ;
      }else{
	double scl ;
	vec2D crsp =p-c ;
	vec2D lcv  =v-c ;
	vec2D onlcv=lcv ; scl=crsp.dot(lcv)/lcv.length() ; onlcv=(onlcv.normalize())*scl ;
	if(      scl<0.0                     ) onlcv=c ;
	else if( lcv.length()<onlcv.length() ) onlcv=v ;
	else                                   onlcv+=c ;
	vec2D lvs  =s-v ; crsp=p-v ;
	vec2D onlvs=lvs ; scl=crsp.dot(lvs)/lvs.length() ; onlvs=(onlvs.normalize())*scl ;
	if(      scl<0.0                     ) onlvs=v ;
	else if( lvs.length()<onlvs.length() ) onlvs=s ;
	else                                   onlvs+=v ;
	vec2D lsc  =c-s ; crsp=p-s ;
	vec2D onlsc=lsc ; scl=crsp.dot(lsc)/lsc.length() ; onlsc=(onlsc.normalize())*scl ;
	if(      scl<0.0                     ) onlsc=s ;
	else if( lsc.length()<onlsc.length() ) onlsc=c ;
	else                                   onlsc+=s ;

	double dplcv=(onlcv-p).length() ;
	double dplvs=(onlvs-p).length() ;
	double dplsc=(onlsc-p).length() ;

	double mind=dplcv ;
	vec2D  curp=onlcv ;
	if( mind>dplvs ){
	  mind=dplvs ;
	  curp=onlvs ;
	}
	if( mind>dplsc ){
	  mind=dplsc ;
	  curp=onlsc ;
	}

	vecPC=c-curp ;
	vecPV=v-curp ;
	vecPS=s-curp ; 

	crsCV = vecPC.cross(vecPV) ;
	crsVS = vecPV.cross(vecPS) ;
	crsSC = vecPS.cross(vecPC) ;
	
	double sumS = crsCV+crsVS+crsSC ;
	cur_sat = abs(crsSC/(sumS)) ;
	cur_val = abs(crsCV/(sumS)) ;
      }
    }
    
    return(0) ;
  }

  
  int drawColorTriangle(double hue, double sat, double val){
    if( bgmap != null ){
      double th = -hue ;
      
      double ri ;
      if( winW<winH ) ri = cast(double)(winW)/2.0 - trir;
      else            ri = cast(double)(winH)/2.0 - trir;
      
      vec2D  c = {ri*cos(th              ), ri*sin(th              )} ;
      vec2D  v = {ri*cos(th + PI*2.0/3.0 ), ri*sin(th + PI*2.0/3.0 )} ;
      vec2D  s = {ri*cos(th - PI*2.0/3.0 ), ri*sin(th - PI*2.0/3.0 )} ;

      double irate=(ri-2.0)/ri ;
      double orate=(ri+2.0)/ri ;

      for( int y=-cast(int)(ri) ; y<=cast(int)(ri) ; y++ ){
	for( int x=-cast(int)(ri) ; x<=cast(int)(ri) ; x++ ){
	  int r=0,g=0,b=0 ;
	  vec2D p = {cast(double)(x),cast(double)(y)} ;
	  vec2D vecPC, vecPV, vecPS ;
	  vecPC=c*irate-p ;
	  vecPV=v*irate-p ;
	  vecPS=s*irate-p ; 
	  double icrsCV = vecPC.cross(vecPV) ;
	  double icrsVS = vecPV.cross(vecPS) ;
	  double icrsSC = vecPS.cross(vecPC) ;
	  vecPC=c*orate-p ;
	  vecPV=v*orate-p ;
	  vecPS=s*orate-p ; 
	  double ocrsCV = vecPC.cross(vecPV) ;
	  double ocrsVS = vecPV.cross(vecPS) ;
	  double ocrsSC = vecPS.cross(vecPC) ;
	  if( icrsCV>=0 && icrsVS>=0 && icrsSC>=0 ){
	    uint col=pos2col(ri*cos(th), ri*sin(th)) ;
	    double sumS = icrsCV+icrsVS+icrsSC ;
	    double fr = cast(double)((col>>16)&0x000000ff)*icrsVS/(sumS) + 255.0*icrsSC/(sumS) ;
	    double fg = cast(double)((col>>8 )&0x000000ff)*icrsVS/(sumS) + 255.0*icrsSC/(sumS) ;
	    double fb = cast(double)((col    )&0x000000ff)*icrsVS/(sumS) + 255.0*icrsSC/(sumS) ;
	    r=cast(int)(fr)*sample*sample ;
	    g=cast(int)(fg)*sample*sample ;
	    b=cast(int)(fb)*sample*sample ;
	  }else if( !(ocrsCV>=0 && ocrsVS>=0 && ocrsSC>=0) ){
	    b=((bgcircle[(y+winH/2)*winW+(x+winW/2)]    )&0x000000ff)*sample*sample ;
	    g=((bgcircle[(y+winH/2)*winW+(x+winW/2)]>>8 )&0x000000ff)*sample*sample ;
	    r=((bgcircle[(y+winH/2)*winW+(x+winW/2)]>>16)&0x000000ff)*sample*sample ;
	  }else{
	    for( int sy=0 ; sy<sample ; sy++ ){
	      for( int sx=0 ; sx<sample ; sx++ ){
		vec2D o = {cast(double)(x)+sx/cast(double)(sample),cast(double)(y)+sy/cast(double)(sample)} ;
		vecPC=c-o ;
		vecPV=v-o ;
		vecPS=s-o ; 
		double crsCV = vecPC.cross(vecPV) ;
		double crsVS = vecPV.cross(vecPS) ;
		double crsSC = vecPS.cross(vecPC) ;
		if( crsCV>=0 && crsVS>=0 && crsSC>=0 ){ //check inside
		  uint col=pos2col(ri*cos(th), ri*sin(th)) ;
		  double sumS = crsCV+crsVS+crsSC ;
		  double fr = cast(double)((col>>16)&0x000000ff)*crsVS/(sumS) + 255.0*crsSC/(sumS) ;
		  double fg = cast(double)((col>>8 )&0x000000ff)*crsVS/(sumS) + 255.0*crsSC/(sumS) ;
		  double fb = cast(double)((col    )&0x000000ff)*crsVS/(sumS) + 255.0*crsSC/(sumS) ;
		  r+=fr ;
		  g+=fg ;
		  b+=fb ;
		}else{
		  b+=(bgcircle[(y+winH/2)*winW+(x+winW/2)]    )&0x000000ff;
		  g+=(bgcircle[(y+winH/2)*winW+(x+winW/2)]>>8 )&0x000000ff;
		  r+=(bgcircle[(y+winH/2)*winW+(x+winW/2)]>>16)&0x000000ff;
		}
	      }
	    }
	  }
	  r/=sample*sample ;
	  g/=sample*sample ;
	  b/=sample*sample ;
	  bgcletri[(y+winH/2)*winW+(x+winW/2)] = 0xff000000 | (cast(int)(r)&0xff)<<16 | (cast(int)(g)&0xff)<<8 | (cast(int)(b)&0xff) ;
	}
      }
    }
    return(0) ;
  }
  
  
  int fillWheel(){
    if( bgmap != null ){
      double ro, ri ;
      double bro, bri ;
      if( winW<winH ) ro = cast(double)(winW)/2.0 ;
      else            ro = cast(double)(winH)/2.0 ;
      ri = ro - circlew ;
      bro=ro+2.0 ;
      bri=ri-2.0 ;
      for( int y=0 ; y<winH ; y++ ){
	for( int x=0 ; x<winW ; x++ ){
	  int    r=0,g=0,b=0 ;
	  double px = x - cast(double)(winW)/2.0 ;
	  double py = y - cast(double)(winH)/2.0 ;
	  double rl = sqrt(px*px+py*py) ;
	  if( rl<bri || rl>bro ){
	    r=(cast(int)(bg_color    ) & 0x000000ff)*sample*sample  ;
	    g=(cast(int)(bg_color>>8 ) & 0x000000ff)*sample*sample  ;
	    b=(cast(int)(bg_color>>16) & 0x000000ff)*sample*sample  ;
	  }else{
	    for( int sy=0 ; sy<sample ; sy++ ){
	      for( int sx=0 ; sx<sample ; sx++ ){
		double dx = px + sx/cast(double)(sample) ;
		double dy = py + sy/cast(double)(sample) ;
		double drl=sqrt(dx*dx+dy*dy) ;
		if( drl<ri || drl>ro ){
		  r+=cast(int)(bg_color    ) & 0x000000ff ;
		  g+=cast(int)(bg_color>>8 ) & 0x000000ff ;
		  b+=cast(int)(bg_color>>16) & 0x000000ff ;
		}else{
		  uint col = pos2col(dx,dy) ;
		  r+=cast(int)(col    ) & 0x000000ff ;
		  g+=cast(int)(col>>8 ) & 0x000000ff ;
		  b+=cast(int)(col>>16) & 0x000000ff ;
		}
	      }
	    }
	  }
	  r/=sample*sample ;
	  g/=sample*sample ;
	  b/=sample*sample ;
	  uint c = b<<16 | g<<8 | r ;
	  bgcletri[y*winW+x] = c ;
	  bgcircle[y*winW+x] = c ;
	}
      }
    }
    return(0) ;
  }

  int drawCircle(double x0, double y0, double r, uint col){
    int ix0=cast(int)(x0) ;
    int iy0=cast(int)(y0) ;
    int x=cast(int)(r) ;
    int y=0 ;
    int d= -2*cast(int)(r)+3 ;
    while( x>=y ){
      bgmap[(iy0+y)*winW+(ix0+x)]=col ;
      bgmap[(iy0+y)*winW+(ix0-x)]=col ;
      bgmap[(iy0-y)*winW+(ix0+x)]=col ;
      bgmap[(iy0-y)*winW+(ix0-x)]=col ;
      bgmap[(iy0+x)*winW+(ix0+y)]=col ;
      bgmap[(iy0+x)*winW+(ix0-y)]=col ;
      bgmap[(iy0-x)*winW+(ix0+y)]=col ;
      bgmap[(iy0-x)*winW+(ix0-y)]=col ;
      if( d>=0 ){
        x-- ;
        d -= 4*x ;
      }
      y++ ;
      d += 4*y+2 ;
    }
    return(0) ;
  }    

  int drawLine(double x0, double y0, double x1, double y1, uint col){
    int dx, dy ;
    int sx, sy, ex, ey ;
    int v ;
    dx = cast(int)(x1-x0) ;
    dy = cast(int)(y1-y0) ;
    v = (dx*dy)>=0 ? 1 : -1 ;

    dx=ABS(dx) ;
    dy=ABS(dy) ;
    if( dx >= dy ){
      if( x0<x1 ){ sx=cast(int)(x0) ; sy=cast(int)(y0) ; ex=cast(int)(x1) ; ey=cast(int)(y1) ;}
      else       { sx=cast(int)(x1) ; sy=cast(int)(y1) ; ex=cast(int)(x0) ; ey=cast(int)(y0) ;}
      int y=sy ;
      int d=dx/2  ;
      for( int x=sx ; x<=ex ; x++ ){
	if( x>=0 && x<winW && y>=0 && y<winH ){
	  bgmap[y*winW+x]=col ;
	}
	d+=dy ;
	if( d>=dx ){
	  y+=v ;
	  d-=dx ;
	}
      }
    }else{
      if( y0<y1 ){ sx=cast(int)(x0) ; sy=cast(int)(y0) ; ex=cast(int)(x1) ; ey=cast(int)(y1) ;}
      else       { sx=cast(int)(x1) ; sy=cast(int)(y1) ; ex=cast(int)(x0) ; ey=cast(int)(y0) ;}
      int x=sx ;
      int d=dy/2  ;
      for( int y=sy ; y<=ey ; y++ ){
	if( x>=0 && x<winW && y>=0 && y<winH ){
	  bgmap[y*winW+x]=col ;
	}
	d+=dx ;
	if( d>=dy ){
	  x+=v ;
	  d-=dy ;
	}
      }
    }
    return(0) ;
  }
  
  uint pos2col(double x, double y){
    vec2D  p  = {x , y} ;
    double th = acos( p.x/p.length() ) ;
    uint   col=bg_color ;
    double ro, ri ;
    
    if( p.y<0 ){
      if(        th>=0.0      && th<(PI/3.0) ){
	int g = cast(int)(round(        255.0 *  th/(PI/3.0))) ;
	col = 0xffff0000 | (g & 0x000000ff)<<8 ;
      }else if( th>=(PI/3.0) && th<(PI/3.0)*2 ){
	int r = cast(int)(round(255.0 - 255.0 * (th-(PI/3.0)  )/(PI/3.0))) ;
	col = 0xff00ff00 | (r & 0x000000ff)<<16 ;
      }else if( th>=(PI/3.0)*2 && th<(PI/3.0)*3 ){
	int b = cast(int)(round(        255.0 * (th-(PI/3.0)*2)/(PI/3.0))) ;
	col = 0xff00ff00 | (b & 0x000000ff) ;
      }
    }else{
      if(        th>=0.0      && th<(PI/3.0) ){
	int b = cast(int)(round(        255.0 *  th/(PI/3.0))) ;
	col = 0xffff0000 | (b & 0x000000ff) ;
      }else if( th>=(PI/3.0) && th<(PI/3.0)*2 ){
	int r = cast(int)(round(255.0 - 255.0 * (th-(PI/3.0)  )/(PI/3.0))) ;
	col = 0xff0000ff | (r & 0x000000ff)<<16 ;
      }else if( th>=(PI/3.0)*2 && th<(PI/3.0)*3 ){
	int g = cast(int)(round(        255.0 * (th-(PI/3.0)*2)/(PI/3.0))) ;
	col = 0xff0000ff | (g & 0x000000ff)<<8 ;
      }
	
    }
    return(col) ;
  }

  int setRGB(uint col){
    int r = cast(int)(col    )&0x000000ff;
    int g = cast(int)(col>>8 )&0x000000ff;
    int b = cast(int)(col>>16)&0x000000ff;

    if( r==g && g==b ){
      cur_sat = cast(double)(r)/255.0 ;
      cur_val = 1.0 - cur_sat ;
    }else{
      double max=r ; if( max<g ) max=g ; if( max<b ) max=b ;
      int    typ=0 ;
      double min=r ; if( min>g ){min=g;typ=1;} ; if( min>b ){min=b;typ=2;} ;

      cur_sat = min/255.0 ;
      cur_val = 1.0-max/255.0 ;
      switch(typ){
      case 0 : //min=r
	if( g<=b ){
	  cur_hue = -PI*2.0/3.0 -(PI*255.0*cast(double)(g)/(3.0*max*255.0)) ;
	}else{
	  cur_hue =  PI*2.0/3.0 +(PI*255.0*cast(double)(b)/(3.0*max*255.0)) ;
	}
	break ;
      case 1 : //min=g
	if( r<=b ){
	  cur_hue = -PI*2.0/3.0 +(PI*255.0*cast(double)(r)/(3.0*max*255.0)) ;
	}else{
	  cur_hue =             -(PI*255.0*cast(double)(b)/(3.0*max*255.0)) ;
	}
	break ;
      case 2 : //min=b
	if( r<=g ){
	  cur_hue =  PI*2.0/3.0 -(PI*255.0*cast(double)(r)/(3.0*max*255.0)) ;
	}else{
	  cur_hue =              (PI*255.0*cast(double)(g)/(3.0*max*255.0)) ;
	}
	break ;
      default : break ;
      }
    }
    
    RedrawWindow( getHWND(), null, null, RDW_INVALIDATE|RDW_UPDATENOW) ;
    return(0) ;
  }

  uint getRGB(){
    int r,g,b ;
    
    if( (cur_val+cur_sat)==1.0 ){
      r = cast(int)(round(255.0*cur_sat)) ;
      g = r ;
      b = r ;
    }else{
      if( cur_hue>=0.0 ){
	if(       cur_hue>=0.0        && cur_hue<PI/3.0     ){
	  r = cast(int)(round(                255.0*(1.0-cur_val))) ;
	  g = cast(int)(round(255.0*cur_sat + 255.0*cur_hue/(PI/3.0)*(1.0-(cur_sat+cur_val)))) ;
	  b = cast(int)(round(255.0*cur_sat)) ;
	}else if( cur_hue>=PI/3.0     && cur_hue<PI*2.0/3.0 ){
	  r = cast(int)(round(255.0*cur_sat + 255.0*(1.0-(cur_hue-PI/3.0)/(PI/3.0))*(1.0-(cur_sat+cur_val)))) ;
	  g = cast(int)(round(                255.0*(1.0-cur_val))) ;
	  b = cast(int)(round(255.0*cur_sat)) ;
	}else if( cur_hue>=PI*2.0/3.0 && cur_hue<PI         ){
	  r = cast(int)(round(255.0*cur_sat)) ;
	  g = cast(int)(round(                255.0*(1.0-cur_val))) ;
	  b = cast(int)(round(255.0*cur_sat + 255.0*(cur_hue-PI*2.0/3.0)/(PI/3.0)*(1.0-(cur_sat+cur_val)))) ;
	}
      }else{
	if(       cur_hue< 0.0        && cur_hue>=-PI/3.0     ){
	  r = cast(int)(round(                255.0*(1.0-cur_val))) ;
	  g = cast(int)(round(255.0*cur_sat)) ;
	  b = cast(int)(round(255.0*cur_sat + 255.0*-cur_hue/(PI/3.0)*(1.0-(cur_sat+cur_val)))) ;
	}else if( cur_hue<-PI/3.0     && cur_hue>=-PI*2.0/3.0 ){
	  r = cast(int)(round(255.0*cur_sat + 255.0*(1.0-(-cur_hue-PI/3.0)/(PI/3.0))*(1.0-(cur_sat+cur_val)))) ;
	  g = cast(int)(round(255.0*cur_sat)) ;
	  b = cast(int)(round(                255.0*(1.0-cur_val))) ;
	}else if( cur_hue<-PI*2.0/3.0 && cur_hue>=-PI         ){
	  r = cast(int)(round(255.0*cur_sat)) ;
	  g = cast(int)(round(255.0*cur_sat + 255.0*(-cur_hue-PI*2.0/3.0)/(PI/3.0)*(1.0-(cur_sat+cur_val)))) ;
	  b = cast(int)(round(                255.0*(1.0-cur_val))) ;
	}
      }
    }
    uint col = (b&0xff)<<16 | (g&0xff)<<8 | (r&0xff) ;
    return(col) ;
  }
}
