import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.WindowObject;
import win32frame.Panel;
import win32frame.CheckBox;
import win32frame.RadioButton;
import win32frame.PushButton;
import win32frame.StaticControl;
import win32frame.TextBox;
import win32frame.WindowMenu;
import win32frame.ToolTip;
import win32frame.ComboBox;
import win32frame.ListView;
import win32frame.TreeView;
import win32frame.Split;
import win32frame.TrackBar;
import win32frame.Progressbar;
import util.Image;
import std.stdio;
import std.string;
import std.math;
import std.utf;

import FontMan;

enum CMD{
  NONE   = 0x0ff,
  EDTREE = 0x100,
  GDISP  = 0x101,
  CHKBOX = 0x102,
  RADIO  = 0x103,
  COMBO  = 0x104,
  EDTXT  = 0x105,
  TREEV  = 0x106,
  LISTV  = 0x108,
  TRKB   = 0x109
} ;

class SubPanel1 : WindowObject {
  Split         split1    ;
  Split         split2    ;
  Panel         panel     ;
  CheckBox      chkbox    ;
  RadioButton   radio1    ;
  RadioButton   radio2    ;
  PushButton    button1   ;
  StaticControl strctl    ;
  TextBox       textbox   ;
  ToolTip       tooltip   ;
  ComboBox      combobox  ;
  ListView      listview  ;
  TreeView      treeview  ;
  TrackBar      trackbar  ;
  Progressbar   progbar   ;

  HFONT         hfont_v14,hfont_f12;

  int           progcur   ;

  this(string name, string ClassName) {
    winc.style          = 0;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.hIcon          = null ;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    //    winc.hbrBackground  = CreateSolidBrush(GetSysColor(COLOR_BTNFACE)) ;
    winc.lpszMenuName   = null;
    winc.lpszClassName  = toUTFz!(wchar*)(ClassName) ;
    super(name) ;
    this.registwinc() ;
  }

  override LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      DeleteObject(hfont_v14);
      DeleteObject(hfont_f12);
      break ;
      
    case WM_CREATE:
      FontMan font_v14 ; font_v14.init(14,VARIABLE_PITCH) ;
      FontMan font_f12 ; font_f12.init(12,FIXED_PITCH) ;
      
      hfont_v14 = font_v14.createHFONT() ;
      hfont_f12 = font_f12.createHFONT() ;
      
      split1 = new Split("Split1",1) ;
      split1.setStyle( WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE ) ;
      split1.setPosition(0,0,300,200) ;
      split1.setSashPos(212,50,55250) ;
      addObject( split1, cast(HMENU)(CMD.NONE), true, true) ;

      panel = new Panel("Panel1") ;
      panel.setStyle( WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE ) ;
      panel.setPosition(0,0,300,200) ;
      split1.addObject( panel, cast(HMENU)(CMD.NONE)) ;

      button1= new PushButton("EditTree") ;
      button1.setPosition(0,0,100,20) ;
      panel.addObject(button1, cast(HMENU)(CMD.EDTREE)) ;
      button1.setFont(hfont_v14) ;
      tooltip=new ToolTip() ;
      tooltip.setText("ツリーのエディット") ;
      button1.addToolTip( tooltip ) ;
      tooltip.setClientRect() ;
      
      strctl= new StaticControl("スタティック\nコントロール") ;
      strctl.addExStyle(WS_EX_CLIENTEDGE) ;
      strctl.addStyle(SS_CENTER) ;
      strctl.setPosition(100,0,100,40) ;
      panel.addObject(strctl, null) ;

      progbar= new Progressbar("プログレスバー") ;
      progbar.setPosition(0,20,100,20) ;
      panel.addObject(progbar,null) ;
      progbar.setRange(0,100) ;
      progbar.setPos(0) ;

      chkbox  = new CheckBox("Check") ;
      chkbox.setPosition(0,40,100,20) ;
      chkbox.addStyle(BS_AUTOCHECKBOX) ;
      panel.addObject(chkbox, cast(HMENU)(CMD.CHKBOX)) ;
      chkbox.setFont(hfont_f12) ;
      chkbox.setState(BST_CHECKED) ;
      chkbox.setWindowText("Check") ;
      
      radio1= new RadioButton("Radio1") ;
      radio1.setPosition(100,40,100,20) ;
      radio1.addStyle(WS_GROUP) ;
      panel.addObject(radio1, cast(HMENU)(CMD.RADIO)) ;
      radio1.setFont(hfont_f12) ;
      radio2= new RadioButton("Radio2") ;
      radio2.setPosition(100,60,100,20) ;
      panel.addObject(radio2, cast(HMENU)(CMD.RADIO)) ;
      radio2.setFont(hfont_f12) ;
      radio2.setState(BST_CHECKED) ;
      
      combobox = new ComboBox("Combo") ;
      combobox.addStyle(CBS_SORT | CBS_DROPDOWNLIST | WS_VSCROLL) ;
      combobox.setPosition(0,80,100,90) ;
      panel.addObject(combobox, cast(HMENU)(CMD.COMBO)) ;
      combobox.setFont(hfont_v14) ;
      combobox.addString("へっぽこ") ;
      combobox.addString("Heppoko") ;
      combobox.addString("Test1") ;
      combobox.addString("Test2") ;
      combobox.setCursel(0) ;

      textbox= new TextBox("Default Text(デフォルトテキスト)") ;
      textbox.addExStyle(WS_EX_CLIENTEDGE) ;
      textbox.addStyle(WS_BORDER | WS_HSCROLL | WS_VSCROLL | ES_MULTILINE | ES_LEFT | ES_AUTOVSCROLL | ES_AUTOHSCROLL) ;
      textbox.setPosition(0,105,300,80) ;
      panel.addObject(textbox, cast(HMENU)(CMD.EDTXT), true, false) ;
      textbox.setFont(hfont_v14) ;
      //      textbox.setWindowText("ChangeText(更新テキスト)") ;

      trackbar = new TrackBar("trackbar") ;
      //      trackbar.addStyle( TBS_NOTICKS ) ; //TBS_NOTHUMB | TBS_AUTOTICKS | TBS_ENABLESELRANGE | TBS_VERT ) ;
      trackbar.setPosition(0,185,300,20) ;
      panel.addObject(trackbar, cast(HMENU)(CMD.TRKB), true,false) ;
      trackbar.setRange(1,256) ;

      split2 = new Split("Split2",1) ;
      split2.setStyle( WS_CHILD |WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE ) ;
      split2.setPosition(0,240,300,200) ;
      split2.setSashPos(68,50,55250) ;
      split1.addObject( split2, cast(HMENU)(CMD.NONE) ) ;

      treeview = new TreeView("TreeView") ;
      treeview.addExStyle(WS_EX_CLIENTEDGE) ;
      treeview.addStyle( WS_BORDER | WS_VSCROLL | TVS_HASBUTTONS | TVS_HASLINES | TVS_LINESATROOT | TVS_EDITLABELS) ;
      treeview.setPosition(0,190,300,50) ;
      split2.addObject(treeview, cast(HMENU)(CMD.TREEV)) ;
      TVINSERTSTRUCT tvins ;
      HTREEITEM      hTreeItem ;
      tvins.hParent      = cast(HTREEITEM)(TVI_ROOT);
      tvins.hInsertAfter = cast(HTREEITEM)(TVI_SORT);
      tvins.item.mask    = TVIF_TEXT;
      tvins.item.pszText = toUTFz!(wchar*)("ツリービュー") ;
      hTreeItem = treeview.insertItem(&tvins);
      tvins.hParent = hTreeItem ;
      tvins.item.pszText = toUTFz!(wchar*)("Heppoko") ;
      treeview.insertItem(&tvins) ;
      tvins.item.pszText = toUTFz!(wchar*)("実験室") ;
      treeview.insertItem(&tvins) ;

      listview = new ListView("ListView") ;
      listview.addExStyle(WS_EX_CLIENTEDGE) ;
      listview.addStyle(LVS_REPORT) ;
      listview.setPosition(250,130,250,90) ;
      split2.addObject(listview, cast(HMENU)(98)) ;
      LVCOLUMN lvcol ;
      lvcol.mask    = LVCF_FMT | LVCF_TEXT | LVCF_WIDTH | LVCF_SUBITEM ;
      lvcol.fmt     = LVCFMT_LEFT ;
      lvcol.cx      = 100 ;
      lvcol.iSubItem = 0 ;
      lvcol.pszText = toUTFz!(wchar*)("Column1") ;
      listview.insertColumn(0, &lvcol) ;
      lvcol.iSubItem = 1 ;
      lvcol.pszText = toUTFz!(wchar*)("Column2") ;
      listview.insertColumn(1, &lvcol) ;
      LVITEM lvitem ;
      lvitem.mask = LVIF_TEXT ;
      lvitem.pszText  = toUTFz!(wchar*)("へっぽこ") ;
      lvitem.iItem    = 0 ;
      lvitem.iSubItem = 0 ;
      listview.insertItem(&lvitem) ;
      lvitem.pszText  = toUTFz!(wchar*)("実験室") ;
      lvitem.iItem    = 0 ;
      lvitem.iSubItem = 1 ;
      listview.setItem(&lvitem) ;
      lvitem.mask = LVIF_TEXT ;
      lvitem.pszText  = toUTFz!(wchar*)("Heppoko") ;
      lvitem.iItem    = 1 ;
      lvitem.iSubItem = 0 ;
      listview.insertItem(&lvitem) ;
      lvitem.pszText  = toUTFz!(wchar*)("Labo") ;
      lvitem.iItem    = 1 ;
      lvitem.iSubItem = 1 ;
      listview.setItem(&lvitem) ;

      SetTimer(hwnd,1,180,null) ;
      
      break ;
      
    case WM_PAINT:
      PAINTSTRUCT ps;
      HDC hdc;
      hdc = BeginPaint(hwnd, &ps);
      EndPaint(hwnd, &ps);
      break ;

    case WM_TIMER:
      progcur = (progcur+1)%110 ;
      progbar.setPos(progcur) ;
      break ;

    case WM_NOTIFY :
      switch( LOWORD(wp) ){
      case CMD.TREEV :
      	TV_DISPINFO *ptv_disp ;
      	ptv_disp = cast(TV_DISPINFO*)(lp) ;
      	if( ptv_disp.hdr.code == TVN_ENDLABELEDIT){
      	  treeview.setItem(&ptv_disp.item) ;
      	}
      	break ;
      default : break ;
      }
      break ;

    case WM_HSCROLL:
      if( trackbar.getHWND()==cast(HWND)(lp) ){
	printStatusbar( format("TrackBar=%d\n",trackbar.getPos()) ) ;
      }
      break ;

    case WM_COMMAND:
      switch( LOWORD(wp) ){
      case CMD.EDTREE :
	HTREEITEM hItem;
	hItem = treeview.getSelection();
	treeview.editLabel(hItem);
	printStatusbar( format("ツリーエディット(%08x)",cast(int)(wp)) ) ;
	break;
	
      case CMD.COMBO :
	int code=HIWORD(wp) ;
	switch(code){
	case CBN_SELCHANGE :
	  printStatusbar( format("コンボボックスチェンジ(%s)",combobox.getCurLabel()) ) ;
	  break ;
	default : break ;
	}
	break ;
	
      case CMD.EDTXT:
	int code=HIWORD(wp) ;
	switch(code){
	case EN_SETFOCUS :
	  printStatusbar("テキストエディットフォーカスを得た") ;
	  break ;
	case EN_UPDATE :
	  printStatusbar("テキストエディット更新された") ;
	  break ;
	default : break ;
	}
	break ;
	
      default :	break ;
      }
      break;
      
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

}
