import win32.windows;
import std.string;
import std.stdio;
import std.utf;
import win32frame.writek;
import win32frame.WindowObject ;

class Launcher : WindowObject {
  bool  initialized=false ;
  int[] cellw, cellh ;
  int   denomw, denomh ;
  int   addpos       ;
  bool  autodiv      ;

  this(string name) {
    winc.style          = 0;
    this.setStyle(WS_CHILD | WS_TABSTOP | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE) ;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.lpszMenuName   = null;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.hbrBackground  = CreateSolidBrush(GetSysColor(COLOR_3DFACE)) ;
    winc.lpszClassName  = toUTFz!(wchar*)("Launcher") ;
    super(name) ;
    this.registwinc() ;
  }

  int initGrid(int xnum, int ynum, int width, int height, bool autodiv=false){
    if( initialized==true  ) return(-2) ;
    if( xnum<=0 || ynum<=0 ) return(-1) ;
    if( autodiv==false && (width<=0 || height<=0) ) return(-1) ;
    denomw=width *xnum ;
    denomh=height*ynum ;
    cellw.length=xnum ;
    cellh.length=ynum ;
    cellw[]=width ;
    cellh[]=height;
    WOBJ_List.length=xnum*ynum ;
    addpos=0 ;
    this.autodiv=autodiv ;
    initialized=true ;
    return(0) ;
  }

  int initGrid(int[] xlist, int[] ylist, bool autodiv=false){
    if( initialized==true  ) return(-2) ;
    if( xlist.length<=0 || ylist.length<=0 ) return(-1) ;
    denomw=0 ;
    foreach( x ; xlist ){
      if( x<=0 ) return(-3) ;
      denomw+=x ;
    }
    denomh=0 ;
    foreach( y ; ylist ){
      if( y<=0 ) return(-3) ;
      denomh+=y ;
    }
    cellw.length=xlist.length ;
    cellh.length=ylist.length ;
    cellw[]=xlist[] ;
    cellh[]=ylist[] ;
    WOBJ_List.length=xlist.length*ylist.length ;
    addpos=0 ;
    this.autodiv=autodiv ;
    initialized=true ;
    return(0) ;
  }

  override int addObject(WindowObject WObj, HMENU hmenu, bool fitWidth=false, bool fitHeight=false ){
    int  ret=-1 ;
    if( initialized==true ){
      if( addpos < WOBJ_List.length ){
	ret = WObj.create(this.getHWND(), hmenu) ;
	if( WObj.getHWND() != null ){
	  WOBJ_List[addpos] = WObj ;
	  int cw, ch ;
	  if( autodiv==false ){
	    cw = cellw[addpos%cellw.length] ;
	    ch = cellh[addpos/cellw.length] ;
	  }else{
	    cw=winW*cellw[addpos%cellw.length]/denomw ;
	    ch=winH*cellh[addpos/cellw.length]/denomh ;
	  }
	  sendmsg( WM_SIZE, 0, (ch<<16 | cw) );
	  addpos++ ;
	}
      }
    }
    return( ret ) ;
  }

  override int delObject(WindowObject WObj){
    return( 0 ) ;
  }

  int reDraw(){
    sendmsg(WM_SIZE, 0, (winH<<16 | winW) );
    return(0) ;
  }
    
  override LRESULT preproc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch( msg ){
    case WM_CREATE :
      break ;
      
    case WM_MOVE :
      _winX=LOWORD(lp) ;
      _winY=HIWORD(lp) ;
      break ;

    case WM_SIZE :
      int px, py, cw, ch, sw, sh ;
      _winW=LOWORD(lp) ;
      _winH=HIWORD(lp) ;
      
      py = 0 ;
      sh = 0 ;
      for( int y=0 ; y<cast(int)(cellh.length) ; y++, py+=ch ){
	if( autodiv==false ){
	  ch = cellh[y] ;
	}else{
	  ch = (winH-py)*cellh[y]/(denomh-sh) ;
	  sh+= cellh[y];
	}
	px = 0 ;
	sw = 0 ;
	for( int x=0 ; x<cast(int)(cellw.length) ; x++, px+=cw ){
	  if( autodiv==false ){
	    cw = cellw[x] ;
	  }else{
	    cw = (winW-px)*cellw[x]/(denomw-sw) ;
	    sw+= cellw[x] ;
	  }
	  if( WOBJ_List[y*cellw.length+x] !is null ){
	    MoveWindow( WOBJ_List[y*cellw.length+x].getHWND(), px, py, cw, ch, true ) ;
	    RedrawWindow(WOBJ_List[y*cellw.length+x].getHWND(), null, null, RDW_INVALIDATE) ;
	  }
	}
      }
      break ;
      
    case WM_NOTIFY:
    case WM_HSCROLL:
    case WM_VSCROLL:
    case WM_COMMAND:
      sendmsgtoparent(msg, wp,lp) ;
      break ;
      
    default : break ;
    }
    
    return( this.proc(hwnd , msg , wp , lp) ) ;
  }

}
